INSERT INTO llm_token_prices 
    (id, pricing_level, pricing_level_id, provider_id, llm_id, effective_date, prompt_token_cost, prompt_token_cost_unit, completion_token_cost, completion_token_cost_unit, created_by, created_at) 
VALUES 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4', now(), 3e-05, 'USD', 6e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4o', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4o-mini', now(), 1.5e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4o-mini-2024-07-18', now(), 1.5e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4o-2024-05-13', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4o-2024-08-06', now(), 2.5e-06, 'USD', 1e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-turbo-preview', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-0314', now(), 3e-05, 'USD', 6e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-0613', now(), 3e-05, 'USD', 6e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-32k', now(), 6e-05, 'USD', 0.00012, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-32k-0314', now(), 6e-05, 'USD', 0.00012, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-32k-0613', now(), 6e-05, 'USD', 0.00012, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-turbo', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-turbo-2024-04-09', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-1106-preview', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-0125-preview', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-vision-preview', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-4-1106-vision-preview', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-3.5-turbo', now(), 1.5e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-3.5-turbo-0301', now(), 1.5e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-3.5-turbo-0613', now(), 1.5e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-3.5-turbo-1106', now(), 1e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-3.5-turbo-0125', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-3.5-turbo-16k', now(), 3e-06, 'USD', 4e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'gpt-3.5-turbo-16k-0613', now(), 3e-06, 'USD', 4e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'ft:gpt-3.5-turbo', now(), 3e-06, 'USD', 6e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'ft:gpt-4-0613', now(), 3e-05, 'USD', 6e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'ft:gpt-4o-mini-2024-07-18', now(), 3e-07, 'USD', 1.2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-openai', 'ft:davinci-002', now(), 2e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-openai', 'ft:babbage-002', now(), 4e-07, 'USD', 4e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'text-embedding-3-large', now(), 1.3e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'text-embedding-3-small', now(), 2e-08, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'text-embedding-ada-002', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'text-embedding-ada-002-v2', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'text-moderation-stable', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'text-moderation-007', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'text-moderation-latest', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', '256-x-256/dall-e-2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', '512-x-512/dall-e-2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', '1024-x-1024/dall-e-2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'hd/1024-x-1792/dall-e-3', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'hd/1792-x-1024/dall-e-3', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'hd/1024-x-1024/dall-e-3', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'standard/1024-x-1792/dall-e-3', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'standard/1792-x-1024/dall-e-3', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'standard/1024-x-1024/dall-e-3', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'whisper-1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'tts-1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openai', 'tts-1-hd', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/tts-1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/tts-1-hd', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/whisper-1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4o', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/global-standard/gpt-4o-mini', now(), 1.5e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4o-mini', now(), 1.65e-07, 'USD', 6.6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4-turbo-2024-04-09', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4-0125-preview', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4-1106-preview', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4-0613', now(), 3e-05, 'USD', 6e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4-32k-0613', now(), 6e-05, 'USD', 0.00012, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4-32k', now(), 6e-05, 'USD', 0.00012, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4', now(), 3e-05, 'USD', 6e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4-turbo', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-4-turbo-vision-preview', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-35-turbo-16k-0613', now(), 3e-06, 'USD', 4e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-35-turbo-1106', now(), 1e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-35-turbo-0125', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-35-turbo-16k', now(), 3e-06, 'USD', 4e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/gpt-35-turbo', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-openai', 'azure/gpt-3.5-turbo-instruct-0914', now(), 1.5e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-openai', 'azure/gpt-35-turbo-instruct', now(), 1.5e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/mistral-large-latest', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/mistral-large-2402', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/command-r-plus', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/ada', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/text-embedding-ada-002', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/text-embedding-3-large', now(), 1.3e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/text-embedding-3-small', now(), 2e-08, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/standard/1024-x-1024/dall-e-3', now(), 0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/hd/1024-x-1024/dall-e-3', now(), 0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/standard/1024-x-1792/dall-e-3', now(), 0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/standard/1792-x-1024/dall-e-3', now(), 0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/hd/1024-x-1792/dall-e-3', now(), 0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/hd/1792-x-1024/dall-e-3', now(), 0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure', 'azure/standard/1024-x-1024/dall-e-2', now(), 0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure_ai', 'azure_ai/jamba-instruct', now(), 5e-07, 'USD', 7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure_ai', 'azure_ai/mistral-large', now(), 4e-06, 'USD', 1.2e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure_ai', 'azure_ai/mistral-small', now(), 1e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure_ai', 'azure_ai/Meta-Llama-3-70B-Instruct', now(), 1.1e-06, 'USD', 3.7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure_ai', 'azure_ai/Meta-Llama-31-8B-Instruct', now(), 3e-07, 'USD', 6.1e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure_ai', 'azure_ai/Meta-Llama-31-70B-Instruct', now(), 2.68e-06, 'USD', 3.54e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'azure_ai', 'azure_ai/Meta-Llama-31-405B-Instruct', now(), 5.33e-06, 'USD', 1.6e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-openai', 'babbage-002', now(), 4e-07, 'USD', 4e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-openai', 'davinci-002', now(), 2e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-openai', 'gpt-3.5-turbo-instruct', now(), 1.5e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-openai', 'gpt-3.5-turbo-instruct-0914', now(), 1.5e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anthropic', 'claude-instant-1', now(), 1.63e-06, 'USD', 5.51e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-tiny', now(), 2.5e-07, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-small', now(), 1e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-small-latest', now(), 1e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-medium', now(), 2.7e-06, 'USD', 8.1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-medium-latest', now(), 2.7e-06, 'USD', 8.1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-medium-2312', now(), 2.7e-06, 'USD', 8.1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-large-latest', now(), 3e-06, 'USD', 9e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-large-2402', now(), 4e-06, 'USD', 1.2e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-large-2407', now(), 3e-06, 'USD', 9e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/open-mistral-7b', now(), 2.5e-07, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/open-mixtral-8x7b', now(), 7e-07, 'USD', 7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/open-mixtral-8x22b', now(), 2e-06, 'USD', 6e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/codestral-latest', now(), 1e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/codestral-2405', now(), 1e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/open-mistral-nemo', now(), 3e-07, 'USD', 3e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/open-mistral-nemo-2407', now(), 3e-07, 'USD', 3e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/open-codestral-mamba', now(), 2.5e-07, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/codestral-mamba-latest', now(), 2.5e-07, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'mistral', 'mistral/mistral-embed', now(), 1e-07, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepseek', 'deepseek-chat', now(), 1.4e-07, 'USD', 2.8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'codestral', 'codestral/codestral-latest', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'codestral', 'codestral/codestral-2405', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-codestral', 'text-completion-codestral/codestral-latest', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'text-completion-codestral', 'text-completion-codestral/codestral-2405', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepseek', 'deepseek-coder', now(), 1.4e-07, 'USD', 2.8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/llama2-70b-4096', now(), 7e-07, 'USD', 8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/llama3-8b-8192', now(), 5e-08, 'USD', 8e-08, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/llama3-70b-8192', now(), 5.9e-07, 'USD', 7.9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/llama-3.1-8b-instant', now(), 5.9e-07, 'USD', 7.9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/llama-3.1-70b-versatile', now(), 5.9e-07, 'USD', 7.9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/llama-3.1-405b-reasoning', now(), 5.9e-07, 'USD', 7.9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/mixtral-8x7b-32768', now(), 2.4e-07, 'USD', 2.4e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/gemma-7b-it', now(), 7e-08, 'USD', 7e-08, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/llama3-groq-70b-8192-tool-use-preview', now(), 8.9e-07, 'USD', 8.9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'groq', 'groq/llama3-groq-8b-8192-tool-use-preview', now(), 1.9e-07, 'USD', 1.9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'friendliai', 'friendliai/mixtral-8x7b-instruct-v0-1', now(), 4e-07, 'USD', 4e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'friendliai', 'friendliai/meta-llama-3-8b-instruct', now(), 1e-07, 'USD', 1e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'friendliai', 'friendliai/meta-llama-3-70b-instruct', now(), 8e-07, 'USD', 8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anthropic', 'claude-instant-1.2', now(), 1.63e-07, 'USD', 5.51e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anthropic', 'claude-2', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anthropic', 'claude-2.1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anthropic', 'claude-3-haiku-20240307', now(), 2.5e-07, 'USD', 1.25e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anthropic', 'claude-3-opus-20240229', now(), 1.5e-05, 'USD', 7.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anthropic', 'claude-3-sonnet-20240229', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anthropic', 'claude-3-5-sonnet-20240620', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-text-models', 'text-bison', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-text-models', 'text-bison@001', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-text-models', 'text-bison@002', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-text-models', 'text-bison32k', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-text-models', 'text-bison32k@002', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-text-models', 'text-unicorn', now(), 1e-05, 'USD', 2.8e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-text-models', 'text-unicorn@001', now(), 1e-05, 'USD', 2.8e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-chat-models', 'chat-bison', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-chat-models', 'chat-bison@001', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-chat-models', 'chat-bison@002', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-chat-models', 'chat-bison-32k', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-chat-models', 'chat-bison-32k@002', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-text-models', 'code-bison', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-text-models', 'code-bison@001', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-text-models', 'code-bison@002', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-text-models', 'code-bison32k', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-text-models', 'code-bison-32k@002', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-text-models', 'code-gecko@001', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-text-models', 'code-gecko@002', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-text-models', 'code-gecko', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-text-models', 'code-gecko-latest', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-chat-models', 'codechat-bison@latest', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-chat-models', 'codechat-bison', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-chat-models', 'codechat-bison@001', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-chat-models', 'codechat-bison@002', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-chat-models', 'codechat-bison-32k', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-code-chat-models', 'codechat-bison-32k@002', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-pro', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.0-pro', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.0-pro-001', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.0-ultra', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.0-ultra-001', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.0-pro-002', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.5-pro', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.5-pro-001', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.5-pro-preview-0514', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.5-pro-preview-0215', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.5-pro-preview-0409', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.5-flash', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.5-flash-001', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-1.5-flash-preview-0514', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'gemini-experimental', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-vision-models', 'gemini-pro-vision', now(), 2.5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-vision-models', 'gemini-1.0-pro-vision', now(), 2.5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-vision-models', 'gemini-1.0-pro-vision-001', now(), 2.5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'medlm-medium', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-language-models', 'medlm-large', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-anthropic_models', 'vertex_ai/claude-3-sonnet@20240229', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-anthropic_models', 'vertex_ai/claude-3-5-sonnet@20240620', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-anthropic_models', 'vertex_ai/claude-3-haiku@20240307', now(), 2.5e-07, 'USD', 1.25e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-anthropic_models', 'vertex_ai/claude-3-opus@20240229', now(), 1.5e-05, 'USD', 7.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-llama_models', 'vertex_ai/meta/llama3-405b-instruct-maas', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-mistral_models', 'vertex_ai/mistral-large@latest', now(), 3e-06, 'USD', 9e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-mistral_models', 'vertex_ai/mistral-large@2407', now(), 3e-06, 'USD', 9e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-mistral_models', 'vertex_ai/mistral-nemo@latest', now(), 3e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-mistral_models', 'vertex_ai/mistral-nemo@2407', now(), 3e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-mistral_models', 'vertex_ai/codestral@latest', now(), 1e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-mistral_models', 'vertex_ai/codestral@2405', now(), 1e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-image-models', 'vertex_ai/imagegeneration@006', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-embedding-models', 'text-embedding-004', now(), 6.25e-09, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-embedding-models', 'text-multilingual-embedding-002', now(), 6.25e-09, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-embedding-models', 'textembedding-gecko', now(), 6.25e-09, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-embedding-models', 'textembedding-gecko-multilingual', now(), 6.25e-09, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-embedding-models', 'textembedding-gecko-multilingual@001', now(), 6.25e-09, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-embedding-models', 'textembedding-gecko@001', now(), 6.25e-09, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-embedding-models', 'textembedding-gecko@003', now(), 6.25e-09, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-embedding-models', 'text-embedding-preview-0409', now(), 6.25e-09, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'vertex_ai-embedding-models', 'text-multilingual-embedding-preview-0409', now(), 6.25e-09, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'palm', 'palm/chat-bison', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'palm', 'palm/chat-bison-001', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'palm', 'palm/text-bison', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'palm', 'palm/text-bison-001', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'palm', 'palm/text-bison-safety-off', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'palm', 'palm/text-bison-safety-recitation-off', now(), 1.25e-07, 'USD', 1.25e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'gemini', 'gemini/gemini-1.5-flash', now(), 3.5e-07, 'USD', 1.05e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'gemini', 'gemini/gemini-1.5-flash-latest', now(), 3.5e-07, 'USD', 1.05e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'gemini', 'gemini/gemini-pro', now(), 3.5e-07, 'USD', 1.05e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'gemini', 'gemini/gemini-1.5-pro', now(), 3.5e-06, 'USD', 1.05e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'gemini', 'gemini/gemini-1.5-pro-exp-0801', now(), 3.5e-06, 'USD', 1.05e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'gemini', 'gemini/gemini-1.5-pro-latest', now(), 3.5e-06, 'USD', 1.05e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'gemini', 'gemini/gemini-pro-vision', now(), 3.5e-07, 'USD', 1.05e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'gemini', 'gemini/gemini-gemma-2-27b-it', now(), 3.5e-07, 'USD', 1.05e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'gemini', 'gemini/gemini-gemma-2-9b-it', now(), 3.5e-07, 'USD', 1.05e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cohere_chat', 'command-r', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cohere_chat', 'command-light', now(), 1.5e-05, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cohere_chat', 'command-r-plus', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cohere', 'command-nightly', now(), 1.5e-05, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cohere', 'command', now(), 1.5e-05, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cohere', 'command-medium-beta', now(), 1.5e-05, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cohere', 'command-xlarge-beta', now(), 1.5e-05, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-2-13b', now(), 1e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-2-13b-chat', now(), 1e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-2-70b', now(), 6.5e-07, 'USD', 2.75e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-2-70b-chat', now(), 6.5e-07, 'USD', 2.75e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-2-7b', now(), 5e-08, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-2-7b-chat', now(), 5e-08, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-3-70b', now(), 6.5e-07, 'USD', 2.75e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-3-70b-instruct', now(), 6.5e-07, 'USD', 2.75e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-3-8b', now(), 5e-08, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/meta/llama-3-8b-instruct', now(), 5e-08, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/mistralai/mistral-7b-v0.1', now(), 5e-08, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/mistralai/mistral-7b-instruct-v0.2', now(), 5e-08, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'replicate', 'replicate/mistralai/mixtral-8x7b-instruct-v0.1', now(), 3e-07, 'USD', 1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/deepseek/deepseek-coder', now(), 1.4e-07, 'USD', 2.8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/microsoft/wizardlm-2-8x22b:nitro', now(), 1e-06, 'USD', 1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/google/gemini-pro-1.5', now(), 2.5e-06, 'USD', 7.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/mistralai/mixtral-8x22b-instruct', now(), 6.5e-07, 'USD', 6.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/cohere/command-r-plus', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/databricks/dbrx-instruct', now(), 6e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/anthropic/claude-3-haiku', now(), 2.5e-07, 'USD', 1.25e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/anthropic/claude-3-haiku-20240307', now(), 2.5e-07, 'USD', 1.25e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/anthropic/claude-3.5-sonnet', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/anthropic/claude-3-sonnet', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/mistralai/mistral-large', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/cognitivecomputations/dolphin-mixtral-8x7b', now(), 5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/google/gemini-pro-vision', now(), 1.25e-07, 'USD', 3.75e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/fireworks/firellava-13b', now(), 2e-07, 'USD', 2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/meta-llama/llama-3-8b-instruct:free', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/meta-llama/llama-3-8b-instruct:extended', now(), 2.25e-07, 'USD', 2.25e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/meta-llama/llama-3-70b-instruct:nitro', now(), 9e-07, 'USD', 9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/meta-llama/llama-3-70b-instruct', now(), 5.9e-07, 'USD', 7.9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/openai/gpt-4o', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/openai/gpt-4o-2024-05-13', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/openai/gpt-4-vision-preview', now(), 1e-05, 'USD', 3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/openai/gpt-3.5-turbo', now(), 1.5e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/openai/gpt-3.5-turbo-16k', now(), 3e-06, 'USD', 4e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/openai/gpt-4', now(), 3e-05, 'USD', 6e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/anthropic/claude-instant-v1', now(), 1.63e-06, 'USD', 5.51e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/anthropic/claude-2', now(), 1.102e-05, 'USD', 3.268e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/anthropic/claude-3-opus', now(), 1.5e-05, 'USD', 7.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/google/palm-2-chat-bison', now(), 5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/google/palm-2-codechat-bison', now(), 5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/meta-llama/llama-2-13b-chat', now(), 2e-07, 'USD', 2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/meta-llama/llama-2-70b-chat', now(), 1.5e-06, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/meta-llama/codellama-34b-instruct', now(), 5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/nousresearch/nous-hermes-llama2-13b', now(), 2e-07, 'USD', 2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/mancer/weaver', now(), 5.625e-06, 'USD', 5.625e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/gryphe/mythomax-l2-13b', now(), 1.875e-06, 'USD', 1.875e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/jondurbin/airoboros-l2-70b-2.1', now(), 1.3875e-05, 'USD', 1.3875e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/undi95/remm-slerp-l2-13b', now(), 1.875e-06, 'USD', 1.875e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/pygmalionai/mythalion-13b', now(), 1.875e-06, 'USD', 1.875e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/mistralai/mistral-7b-instruct', now(), 1.3e-07, 'USD', 1.3e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'openrouter', 'openrouter/mistralai/mistral-7b-instruct:free', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ai21', 'j2-ultra', now(), 1.5e-05, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ai21', 'j2-mid', now(), 1e-05, 'USD', 1e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ai21', 'j2-light', now(), 3e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'nlp_cloud', 'dolphin', now(), 5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'nlp_cloud', 'chatdolphin', now(), 5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'aleph_alpha', 'luminous-base', now(), 3e-05, 'USD', 3.3e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'aleph_alpha', 'luminous-base-control', now(), 3.75e-05, 'USD', 4.125e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'aleph_alpha', 'luminous-extended', now(), 4.5e-05, 'USD', 4.95e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'aleph_alpha', 'luminous-extended-control', now(), 5.625e-05, 'USD', 6.1875e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'aleph_alpha', 'luminous-supreme', now(), 0.000175, 'USD', 0.0001925, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'aleph_alpha', 'luminous-supreme-control', now(), 0.00021875, 'USD', 0.000240625, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'ai21.j2-mid-v1', now(), 1.25e-05, 'USD', 1.25e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'ai21.j2-ultra-v1', now(), 1.88e-05, 'USD', 1.88e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'ai21.jamba-instruct-v1:0', now(), 5e-07, 'USD', 7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'amazon.titan-text-lite-v1', now(), 3e-07, 'USD', 4e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'amazon.titan-text-express-v1', now(), 1.3e-06, 'USD', 1.7e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'amazon.titan-embed-text-v1', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'amazon.titan-embed-text-v2:0', now(), 2e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'mistral.mistral-7b-instruct-v0:2', now(), 1.5e-07, 'USD', 2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'mistral.mixtral-8x7b-instruct-v0:1', now(), 4.5e-07, 'USD', 7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'mistral.mistral-large-2402-v1:0', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'mistral.mistral-large-2407-v1:0', now(), 3e-06, 'USD', 9e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/mistral.mixtral-8x7b-instruct-v0:1', now(), 4.5e-07, 'USD', 7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/mistral.mixtral-8x7b-instruct-v0:1', now(), 4.5e-07, 'USD', 7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-west-3/mistral.mixtral-8x7b-instruct-v0:1', now(), 5.9e-07, 'USD', 9.1e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/mistral.mistral-7b-instruct-v0:2', now(), 1.5e-07, 'USD', 2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/mistral.mistral-7b-instruct-v0:2', now(), 1.5e-07, 'USD', 2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-west-3/mistral.mistral-7b-instruct-v0:2', now(), 2e-07, 'USD', 2.6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/mistral.mistral-large-2402-v1:0', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/mistral.mistral-large-2402-v1:0', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-west-3/mistral.mistral-large-2402-v1:0', now(), 1.04e-05, 'USD', 3.12e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'anthropic.claude-3-sonnet-20240229-v1:0', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'anthropic.claude-3-5-sonnet-20240620-v1:0', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'anthropic.claude-3-haiku-20240307-v1:0', now(), 2.5e-07, 'USD', 1.25e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'anthropic.claude-3-opus-20240229-v1:0', now(), 1.5e-05, 'USD', 7.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'anthropic.claude-v1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/anthropic.claude-v1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/anthropic.claude-v1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/anthropic.claude-v1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/1-month-commitment/anthropic.claude-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/6-month-commitment/anthropic.claude-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/anthropic.claude-v1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/1-month-commitment/anthropic.claude-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/6-month-commitment/anthropic.claude-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/1-month-commitment/anthropic.claude-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/6-month-commitment/anthropic.claude-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/1-month-commitment/anthropic.claude-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/6-month-commitment/anthropic.claude-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'anthropic.claude-v2', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/anthropic.claude-v2', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/anthropic.claude-v2', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/anthropic.claude-v2', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/1-month-commitment/anthropic.claude-v2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/6-month-commitment/anthropic.claude-v2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/anthropic.claude-v2', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/1-month-commitment/anthropic.claude-v2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/6-month-commitment/anthropic.claude-v2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/1-month-commitment/anthropic.claude-v2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/6-month-commitment/anthropic.claude-v2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/1-month-commitment/anthropic.claude-v2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/6-month-commitment/anthropic.claude-v2', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'anthropic.claude-v2:1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/anthropic.claude-v2:1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/anthropic.claude-v2:1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/anthropic.claude-v2:1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/1-month-commitment/anthropic.claude-v2:1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/6-month-commitment/anthropic.claude-v2:1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/anthropic.claude-v2:1', now(), 8e-06, 'USD', 2.4e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/1-month-commitment/anthropic.claude-v2:1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/6-month-commitment/anthropic.claude-v2:1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/1-month-commitment/anthropic.claude-v2:1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/6-month-commitment/anthropic.claude-v2:1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/1-month-commitment/anthropic.claude-v2:1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/6-month-commitment/anthropic.claude-v2:1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'anthropic.claude-instant-v1', now(), 1.63e-06, 'USD', 5.51e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/anthropic.claude-instant-v1', now(), 8e-07, 'USD', 2.4e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/1-month-commitment/anthropic.claude-instant-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-east-1/6-month-commitment/anthropic.claude-instant-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/1-month-commitment/anthropic.claude-instant-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/6-month-commitment/anthropic.claude-instant-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/us-west-2/anthropic.claude-instant-v1', now(), 8e-07, 'USD', 2.4e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/anthropic.claude-instant-v1', now(), 2.23e-06, 'USD', 7.55e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/1-month-commitment/anthropic.claude-instant-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/ap-northeast-1/6-month-commitment/anthropic.claude-instant-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/anthropic.claude-instant-v1', now(), 2.48e-06, 'USD', 8.38e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/1-month-commitment/anthropic.claude-instant-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/eu-central-1/6-month-commitment/anthropic.claude-instant-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'cohere.command-text-v14', now(), 1.5e-06, 'USD', 2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/*/1-month-commitment/cohere.command-text-v14', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/*/6-month-commitment/cohere.command-text-v14', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'cohere.command-light-text-v14', now(), 3e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/*/1-month-commitment/cohere.command-light-text-v14', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'bedrock/*/6-month-commitment/cohere.command-light-text-v14', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'cohere.command-r-plus-v1:0', now(), 3e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'cohere.command-r-v1:0', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'cohere.embed-english-v3', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'cohere.embed-multilingual-v3', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'meta.llama2-13b-chat-v1', now(), 7.5e-07, 'USD', 1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'meta.llama2-70b-chat-v1', now(), 1.95e-06, 'USD', 2.56e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'meta.llama3-8b-instruct-v1:0', now(), 4e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'meta.llama3-70b-instruct-v1:0', now(), 2.65e-06, 'USD', 3.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'meta.llama3-1-8b-instruct-v1:0', now(), 4e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'meta.llama3-1-70b-instruct-v1:0', now(), 2.65e-06, 'USD', 3.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'meta.llama3-1-405b-instruct-v1:0', now(), 5.32e-06, 'USD', 1.6e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', '512-x-512/50-steps/stability.stable-diffusion-xl-v0', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', '512-x-512/max-steps/stability.stable-diffusion-xl-v0', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'max-x-max/50-steps/stability.stable-diffusion-xl-v0', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', 'max-x-max/max-steps/stability.stable-diffusion-xl-v0', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', '1024-x-1024/50-steps/stability.stable-diffusion-xl-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'bedrock', '1024-x-1024/max-steps/stability.stable-diffusion-xl-v1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'sagemaker', 'sagemaker/meta-textgeneration-llama-2-7b', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'sagemaker', 'sagemaker/meta-textgeneration-llama-2-7b-f', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'sagemaker', 'sagemaker/meta-textgeneration-llama-2-13b', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'sagemaker', 'sagemaker/meta-textgeneration-llama-2-13b-f', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'sagemaker', 'sagemaker/meta-textgeneration-llama-2-70b', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'sagemaker', 'sagemaker/meta-textgeneration-llama-2-70b-b-f', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'together_ai', 'together-ai-up-to-4b', now(), 1e-07, 'USD', 1e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'together_ai', 'together-ai-4.1b-8b', now(), 2e-07, 'USD', 2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'together_ai', 'together-ai-8.1b-21b', now(), 3e-07, 'USD', 3e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'together_ai', 'together-ai-21.1b-41b', now(), 8e-07, 'USD', 8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'together_ai', 'together-ai-41.1b-80b', now(), 9e-07, 'USD', 9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'together_ai', 'together-ai-81.1b-110b', now(), 1.8e-06, 'USD', 1.8e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'together_ai', 'together_ai/mistralai/Mixtral-8x7B-Instruct-v0.1', now(), 6e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'together_ai', 'together_ai/mistralai/Mistral-7B-Instruct-v0.1', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'together_ai', 'together_ai/togethercomputer/CodeLlama-34b-Instruct', now(), 0, 'USD', 0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/codegemma', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/codegeex4', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/deepseek-coder-v2-instruct', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/deepseek-coder-v2-base', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/deepseek-coder-v2-lite-instruct', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/deepseek-coder-v2-lite-base', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/internlm2_5-20b-chat', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/llama2', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/llama2:13b', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/llama2:70b', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/llama2-uncensored', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/llama3', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/llama3:70b', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/llama3.1', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/mistral-large-instruct-2407', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/mistral', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/mistral-7B-Instruct-v0.1', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/mistral-7B-Instruct-v0.2', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/mixtral-8x7B-Instruct-v0.1', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/mixtral-8x22B-Instruct-v0.1', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/codellama', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/orca-mini', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'ollama', 'ollama/vicuna', now(), 0.0, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/lizpreciatior/lzlv_70b_fp16_hf', now(), 7e-07, 'USD', 9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/Gryphe/MythoMax-L2-13b', now(), 2.2e-07, 'USD', 2.2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/mistralai/Mistral-7B-Instruct-v0.1', now(), 1.3e-07, 'USD', 1.3e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/meta-llama/Llama-2-70b-chat-hf', now(), 7e-07, 'USD', 9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/cognitivecomputations/dolphin-2.6-mixtral-8x7b', now(), 2.7e-07, 'USD', 2.7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/codellama/CodeLlama-34b-Instruct-hf', now(), 6e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/deepinfra/mixtral', now(), 2.7e-07, 'USD', 2.7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/Phind/Phind-CodeLlama-34B-v2', now(), 6e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/mistralai/Mixtral-8x7B-Instruct-v0.1', now(), 2.7e-07, 'USD', 2.7e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/deepinfra/airoboros-70b', now(), 7e-07, 'USD', 9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/01-ai/Yi-34B-Chat', now(), 6e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/01-ai/Yi-6B-200K', now(), 1.3e-07, 'USD', 1.3e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/jondurbin/airoboros-l2-70b-gpt4-1.4.1', now(), 7e-07, 'USD', 9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/meta-llama/Llama-2-13b-chat-hf', now(), 2.2e-07, 'USD', 2.2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/amazon/MistralLite', now(), 2e-07, 'USD', 2e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/meta-llama/Llama-2-7b-chat-hf', now(), 1.3e-07, 'USD', 1.3e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/meta-llama/Meta-Llama-3-8B-Instruct', now(), 8e-08, 'USD', 8e-08, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/meta-llama/Meta-Llama-3-70B-Instruct', now(), 5.9e-07, 'USD', 7.9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/01-ai/Yi-34B-200K', now(), 6e-07, 'USD', 6e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'deepinfra', 'deepinfra/openchat/openchat_3.5', now(), 1.3e-07, 'USD', 1.3e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/codellama-34b-instruct', now(), 3.5e-07, 'USD', 1.4e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/codellama-70b-instruct', now(), 7e-07, 'USD', 2.8e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/pplx-7b-chat', now(), 7e-08, 'USD', 2.8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/pplx-70b-chat', now(), 7e-07, 'USD', 2.8e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/pplx-7b-online', now(), 0.0, 'USD', 2.8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/pplx-70b-online', now(), 0.0, 'USD', 2.8e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/llama-2-70b-chat', now(), 7e-07, 'USD', 2.8e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/mistral-7b-instruct', now(), 7e-08, 'USD', 2.8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/mixtral-8x7b-instruct', now(), 7e-08, 'USD', 2.8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/sonar-small-chat', now(), 7e-08, 'USD', 2.8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/sonar-small-online', now(), 0, 'USD', 2.8e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/sonar-medium-chat', now(), 6e-07, 'USD', 1.8e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'perplexity', 'perplexity/sonar-medium-online', now(), 0, 'USD', 1.8e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'fireworks_ai', 'fireworks_ai/accounts/fireworks/models/firefunction-v2', now(), 9e-07, 'USD', 9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'fireworks_ai', 'fireworks_ai/accounts/fireworks/models/mixtral-8x22b-instruct-hf', now(), 1.2e-06, 'USD', 1.2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'fireworks_ai', 'fireworks_ai/accounts/fireworks/models/qwen2-72b-instruct', now(), 9e-07, 'USD', 9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'fireworks_ai', 'fireworks_ai/accounts/fireworks/models/yi-large', now(), 3e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'fireworks_ai', 'fireworks_ai/accounts/fireworks/models/deepseek-coder-v2-instruct', now(), 1.2e-06, 'USD', 1.2e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/mistralai/Mistral-7B-Instruct-v0.1', now(), 1.5e-07, 'USD', 1.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/mistralai/Mixtral-8x7B-Instruct-v0.1', now(), 1.5e-07, 'USD', 1.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/mistralai/Mixtral-8x22B-Instruct-v0.1', now(), 9e-07, 'USD', 9e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/HuggingFaceH4/zephyr-7b-beta', now(), 1.5e-07, 'USD', 1.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/google/gemma-7b-it', now(), 1.5e-07, 'USD', 1.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/meta-llama/Llama-2-7b-chat-hf', now(), 1.5e-07, 'USD', 1.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/meta-llama/Llama-2-13b-chat-hf', now(), 2.5e-07, 'USD', 2.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/meta-llama/Llama-2-70b-chat-hf', now(), 1e-06, 'USD', 1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/codellama/CodeLlama-34b-Instruct-hf', now(), 1e-06, 'USD', 1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/codellama/CodeLlama-70b-Instruct-hf', now(), 1e-06, 'USD', 1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/meta-llama/Meta-Llama-3-8B-Instruct', now(), 1.5e-07, 'USD', 1.5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'anyscale', 'anyscale/meta-llama/Meta-Llama-3-70B-Instruct', now(), 1e-06, 'USD', 1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cloudflare', 'cloudflare/@cf/meta/llama-2-7b-chat-fp16', now(), 1.923e-06, 'USD', 1.923e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cloudflare', 'cloudflare/@cf/meta/llama-2-7b-chat-int8', now(), 1.923e-06, 'USD', 1.923e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cloudflare', 'cloudflare/@cf/mistral/mistral-7b-instruct-v0.1', now(), 1.923e-06, 'USD', 1.923e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'cloudflare', 'cloudflare/@hf/thebloke/codellama-7b-instruct-awq', now(), 1.923e-06, 'USD', 1.923e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'voyage', 'voyage/voyage-01', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'voyage', 'voyage/voyage-lite-01', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'voyage', 'voyage/voyage-large-2', now(), 1.2e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'voyage', 'voyage/voyage-law-2', now(), 1.2e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'voyage', 'voyage/voyage-code-2', now(), 1.2e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'voyage', 'voyage/voyage-2', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'voyage', 'voyage/voyage-lite-02-instruct', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'databricks', 'databricks/databricks-meta-llama-3-1-405b-instruct', now(), 5e-06, 'USD', 1.5e-05, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'databricks', 'databricks/databricks-meta-llama-3-1-70b-instruct', now(), 1e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'databricks', 'databricks/databricks-dbrx-instruct', now(), 7.5e-07, 'USD', 2.25e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'databricks', 'databricks/databricks-meta-llama-3-70b-instruct', now(), 1e-06, 'USD', 3e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'databricks', 'databricks/databricks-llama-2-70b-chat', now(), 5e-07, 'USD', 1.5e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'databricks', 'databricks/databricks-mixtral-8x7b-instruct', now(), 5e-07, 'USD', 1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'databricks', 'databricks/databricks-mpt-30b-instruct', now(), 1e-06, 'USD', 1e-06, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'databricks', 'databricks/databricks-mpt-7b-instruct', now(), 5e-07, 'USD', 5e-07, 'USD', 'system', now()), 
    (generateUUIDv4(), 'default', 'default', 'databricks', 'databricks/databricks-bge-large-en', now(), 1e-07, 'USD', 0.0, 'USD', 'system', now())
